package com.fjhx.flow.service.flow.impl;

import cn.hutool.core.util.StrUtil;
import cn.hutool.extra.spring.SpringUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fjhx.flow.core.FlowBean;
import com.fjhx.flow.core.FlowDelegate;
import com.fjhx.flow.core.FlowThreadLocalUtil;
import com.fjhx.flow.entity.flow.dto.InitiateDto;
import com.fjhx.flow.entity.flow.po.FlowDefinition;
import com.fjhx.flow.entity.flow.po.FlowDefinitionNode;
import com.fjhx.flow.entity.flow.po.FlowExample;
import com.fjhx.flow.entity.flow.po.FlowExampleDetail;
import com.fjhx.flow.enums.FlowStatus;
import com.fjhx.flow.enums.HandleType;
import com.fjhx.flow.enums.NodeTypeEnum;
import com.fjhx.flow.service.flow.*;
import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.exception.ServiceException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Slf4j
@Service
public class FlowProcessServiceImpl implements FlowProcessService {

    @Autowired
    private FlowDefinitionService flowDefinitionService;

    @Autowired
    private FlowDefinitionNodeService flowDefinitionNodeService;

    @Autowired
    private FlowExampleService flowExampleService;

    @Autowired
    private FlowExampleDetailService flowExampleDetailService;

    @Autowired
    private PlatformTransactionManager platformTransactionManager;

    @Autowired
    private TransactionDefinition transactionDefinition;

    @Override
    public void initiate(InitiateDto dto) {

        // 找到代码定义的流程bean
        Class<? extends FlowDelegate> flowDelegateCls = FlowBean.getBean(dto.getFlowKey());

        // 实例化bean
        FlowDelegate flowDelegate;

        flowDelegate = SpringUtil.getBean(flowDelegateCls);
        if (flowDelegate == null) {
            throw new ServiceException("没有把流程bean注册到spring中");
        }

        // 查找流程定义
        FlowDefinition flowDefinition = flowDefinitionService.getOne(q -> q
                .eq(FlowDefinition::getFlowKey, dto.getFlowKey())
                .eq(FlowDefinition::getCurrentVersion, StatusConstant.YES));

        if (flowDefinition == null) {
            throw new ServiceException("没有找到可用流程");
        }

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService
                .list(q -> q.eq(FlowDefinitionNode::getFlowDefinitionId, flowDefinition.getId()));

        // 父级节点map
        Map<Long, List<FlowDefinitionNode>> parentNodeMap = flowDefinitionNodeList.stream()
                .collect(Collectors.groupingBy(FlowDefinitionNode::getParentId));

        // 寻找开始节点
        FlowDefinitionNode startNode = flowDefinitionNodeList.stream()
                .filter(item -> NodeTypeEnum.START.getKey().equals(item.getNodeType()))
                .findFirst().orElse(null);

        if (startNode == null) {
            throw new ServiceException("没有找到开始节点");
        }

        // 生成流程id
        long flowId = IdWorker.getId();

        // 初始化模板参数
        Map<String, Object> templateMap = flowDelegate.initTemplateMap(dto.getData(), dto.getData());

        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(dto.getData());
        FlowThreadLocalUtil.setCurrentData(dto.getData());
        FlowThreadLocalUtil.setFlowId(flowId);

        // 寻找下一节点
        FlowDefinitionNode nextUserNode = getNextUserNode(startNode, parentNodeMap);

        // 流程实例
        FlowExample flowExample = new FlowExample();

        // 流程实例明细列表
        List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();

        // 开启一个事务
        TransactionStatus transaction = platformTransactionManager.getTransaction(transactionDefinition);

        try {

            // 执行开始流程方法
            Long businessId = flowDelegate.start(flowId, dto.getData());
            FlowThreadLocalUtil.setBusinessId(businessId);

            FlowExampleDetail startExampleDetail = new FlowExampleDetail();
            startExampleDetail.setFlowExampleId(flowId);
            startExampleDetail.setFlowDefinitionNodeId(startNode.getId());
            startExampleDetail.setFlowDefinitionNodeType(startNode.getNodeType());
            startExampleDetail.setHandleType(HandleType.NEXT.getKey());
            flowExampleDetailList.add(startExampleDetail);

            // 如果下个节点是结束节点
            if (NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(nextUserNode.getNodeType()))) {
                // 走非默认方法
                String handlingMethod = nextUserNode.getHandlingMethod();
                if (StrUtil.isNotBlank(handlingMethod)) {
                    try {
                        Method method = flowDelegateCls.getMethod(handlingMethod);
                        method.invoke(flowDelegate);
                    } catch (Exception e) {
                        throw new ServiceException("未找到指定处理方法：" + handlingMethod);
                    }
                } else {
                    // 执行流程结束方法
                    flowDelegate.end();
                }

                // 流程已通过
                flowExample.setStatus(FlowStatus.HAVE_PASSED.getKey());

                // 结束流程
                FlowExampleDetail endExampleDetail = new FlowExampleDetail();
                endExampleDetail.setFlowExampleId(flowId);
                endExampleDetail.setFlowDefinitionNodeId(nextUserNode.getId());
                endExampleDetail.setFlowDefinitionNodeType(nextUserNode.getNodeType());
                endExampleDetail.setHandleType(HandleType.OVER.getKey());
                flowExampleDetailList.add(endExampleDetail);
            } else {
                // 流程进行中
                flowExample.setStatus(FlowStatus.IN_PROGRESS.getKey());
            }

            flowExample.setTitle(StrUtil.format(flowDefinition.getTitleTemplate(), templateMap, true));
            flowExample.setDefinitionId(flowDefinition.getId());
            flowExample.setDefintionNodeId(nextUserNode.getId());
            flowExample.setBusinessId(businessId);
            flowExample.setStartData(dto.getData().toJSONString());
            flowExampleService.save(flowExample);
            flowExampleDetailService.saveBatch(flowExampleDetailList);

            // 提交事务
            platformTransactionManager.commit(transaction);
        } catch (Exception e) {
            // 回滚事务
            platformTransactionManager.rollback(transaction);
            log.error("业务错误", e);
            throw new ServiceException("业务代码错误");
        }

    }

    /**
     * 查找下一个用户执行节点
     *
     * @param currentNode   当前节点
     * @param parentNodeMap 父节点map
     * @return 用户节点
     */
    private FlowDefinitionNode getNextUserNode(FlowDefinitionNode currentNode, Map<Long, List<FlowDefinitionNode>> parentNodeMap) {
        // 查找下个节点
        List<FlowDefinitionNode> nextNodeList = parentNodeMap.get(currentNode.getId());
        FlowDefinitionNode nextNode = null;
        // 如果为分支
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(currentNode.getNodeType()))) {
            // 循环分支
            for (FlowDefinitionNode flowDefinitionNode : nextNodeList) {
                // 获取跳转条件
                String jumpCondition = flowDefinitionNode.getJumpCondition();
                // 跳转条件为空表示默认跳转
                if (StrUtil.isBlank(jumpCondition)) {
                    nextNode = flowDefinitionNode;
                }
                // 符合条件
                if (expressionResult(FlowThreadLocalUtil.getTemplateData(), jumpCondition)) {
                    nextNode = flowDefinitionNode;
                    break;
                }
            }
            if (nextNode == null) {
                throw new ServiceException("未识别到分支跳转节点");
            }
        }
        // 如果不为分支
        else {
            if (nextNodeList.size() != 1) {
                throw new ServiceException("流程定义错误");
            }
            // 查找下一个节点方法
            nextNode = nextNodeList.get(0);
        }
        // 如果下一节点为分支，递归找下一节点
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(nextNode.getNodeType()))) {
            return getNextUserNode(nextNode, parentNodeMap);
        }
        return nextNode;
    }

    /**
     * 校验el表达式
     */
    public static boolean expressionResult(Map<String, Object> map, String expression) {
        Expression exp = AviatorEvaluator.compile(expression);
        Object execute = exp.execute(map);
        return Boolean.parseBoolean(String.valueOf(execute));
    }

}
